.floating-bar { background-color: @theme_base_color; border-width: 1px; border-style: solid solid none; border-color: @borders; border-radius: 3px 3px 0 0; }

.floating-bar.left { border-left-style: none; border-top-left-radius: 0; }

.floating-bar.right { border-right-style: none; border-top-right-radius: 0; }

.floating-bar:backdrop { background-color: @theme_unfocused_base_color; border-color: @unfocused_borders; }

#title-box-container { transition: background 200ms ease-out; }

.incognito-mode #title-box-container { background: -gtk-icontheme("user-not-tracked-symbolic") 16px 0/64px 64px no-repeat; }

.incognito-mode #title-box-container:backdrop { background-image: cross-fade(50% -gtk-icontheme("user-not-tracked-symbolic"), image(transparent)); }

.automation-mode headerbar { background-color: #ff9600; color: rgba(0, 0, 0, 0.35); }

.entry_icon_box { margin-left: 2px; margin-right: 5px; }

.entry_icon_box:dir(rtl) { margin-left: 5px; margin-right: 2px; }

.entry_icon_box > button { box-shadow: none; border: transparent; background: transparent; padding: 0px; -gtk-icon-shadow: none; outline-offset: -1px; }

.entry_icon_box > button:hover:not(:backdrop) .entry_icon { opacity: 1; }

.entry_icon { color: @theme_fg_color; margin: 0 4px; opacity: .8; }

.entry_icon:active { color: @theme_selected_bg_color; }

.entry_icon:backdrop { color: @theme_unfocused_fg_color; }

.entry_icon.selected { color: @theme_selected_bg_color; }

.entry_icon.starred { color: orange; }

.bookmarks-row button { opacity: 0.25; }

.bookmarks-row:hover button { opacity: 1; }

.bookmark-tag-widget { padding-left: 8px; background-color: rgba(0, 0, 0, 0.1); border-radius: 3px; }

.bookmark-tag-widget label { padding-left: 8px; padding-right: 8px; }

.bookmark-tag-widget image, .bookmark-tag-widget label, .bookmark-tag-widget button:hover image { color: @theme_fg_color; }

.bookmark-tag-widget-selected { background-color: @theme_selected_bg_color; }

.bookmark-tag-widget-selected image, .bookmark-tag-widget-selected label, .bookmark-tag-widget-selected button:hover image { color: @theme_selected_fg_color; }

#fullscreen-popup { padding: 12px; border-radius: 18px; background: rgba(0, 0, 0, 0.65); color: white; }

.url_progress { border: none; background-color: transparent; background-image: none; padding-top: 0px; padding-bottom: 0px; margin-top: 0px; margin-bottom: 6px; }

.url_progress progress { background-color: @theme_selected_bg_color; background-image: none; border: none; min-height: 2px; }

.url_progress trough { border: none; background-color: transparent; background-image: none; }

.tab-attention { background-image: -gtk-gradient(radial, center center, 0, center center, 0.5, to(@theme_selected_bg_color), to(transparent)); background-size: 6px 6px; background-repeat: no-repeat; background-position: center bottom; }

.tab-arrow-up-attention arrow.up { background-color: @theme_selected_bg_color; }

.tab-arrow-down-attention arrow.down { background-color: @theme_selected_bg_color; }

.page-menu-contents { margin-left: 12px; margin-right: 12px; margin-bottom: 12px; }

.search-entry-occurrences-tag { background: none; color: alpha(@theme_fg_color,0.5); box-shadow: none; border: none; margin: 2px; padding: 2px; }

.search-entry-occurrences-tag:backdrop { color: alpha(@theme_unfocused_fg_color,0.5); }

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > grid { margin-left: 3px; }

dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > grid .title, dzlsuggestionpopover > revealer > box > elastic > scrolledwindow > viewport > list > row > grid .subtitle { margin-left: 6px; }

.download_progress progress { min-height: 2px; }

.download_progress trough { min-width: 14px; }

.pages-list { background: none; }

.draggable.drag-icon { background: white; border: 1px solid black; }

.draggable.drag-row { color: gray; background: alpha(gray,0.2); }

fullscreenbox > flap > dimming, fullscreenbox > flap > outline, fullscreenbox > flap > border { min-height: 0; min-width: 0; background: none; }

fullscreenbox > flap > shadow { min-height: 9px; min-width: 9px; background: linear-gradient(to bottom, alpha(black,0.1), alpha(black,0)); }

.incognito-mode { /*************************** Check and Radio buttons * */ }

.incognito-mode headerbar { background: #141927 linear-gradient(to top, #1a2235, #1e263b); box-shadow: inset 0 1px rgba(231, 234, 243, 0.07); border-color: #06080c; }

.incognito-mode headerbar > clamp { color: #e7eaf3; }

.incognito-mode headerbar:backdrop { background-image: image(#252f49); border-color: #171d2e; box-shadow: inset 0 1px rgba(231, 234, 243, 0.07); }

.incognito-mode headerbar:backdrop > clamp { color: #868d9e; }

.incognito-mode actionbar { background: #252f49; }

.incognito-mode actionbar > revealer > box { border-color: #141927; }

.incognito-mode actionbar:backdrop { background: #252f49; }

.incognito-mode actionbar:backdrop > revealer > box { border-color: #171d2e; }

.incognito-mode headerbar > *, .incognito-mode actionbar > * { color: #e7eaf3; }

.incognito-mode headerbar > *:backdrop, .incognito-mode actionbar > *:backdrop { color: #868d9e; }

.incognito-mode headerbar button, .incognito-mode actionbar button { color: #e7eaf3; outline-color: rgba(231, 234, 243, 0.3); border-color: #141927; border-bottom-color: #06080c; background-image: linear-gradient(to top, #232d46 2px, #252f49); text-shadow: 0 -1px rgba(0, 0, 0, 0.828235); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.828235); box-shadow: inset 0 1px rgba(255, 255, 255, 0.02), 0 1px 2px rgba(0, 0, 0, 0.07); }

.incognito-mode headerbar button.flat, .incognito-mode headerbar button.titlebutton, .incognito-mode actionbar button.flat, .incognito-mode actionbar button.titlebutton { border-color: transparent; background-color: transparent; background-image: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.incognito-mode headerbar button.titlebutton, .incognito-mode actionbar button.titlebutton { text-shadow: 0 -1px rgba(0, 0, 0, 0.828235); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.828235); }

.incognito-mode headerbar button:hover, .incognito-mode actionbar button:hover { color: #e7eaf3; outline-color: rgba(231, 234, 243, 0.3); border-color: #141927; border-bottom-color: #06080c; background-image: linear-gradient(to top, #232d46, #26314c 1px); text-shadow: 0 -1px rgba(0, 0, 0, 0.780235); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.780235); box-shadow: inset 0 1px rgba(255, 255, 255, 0.02), 0 1px 2px rgba(0, 0, 0, 0.07); }

.incognito-mode headerbar button:active, .incognito-mode headerbar button:checked, .incognito-mode actionbar button:active, .incognito-mode actionbar button:checked { color: #e7eaf3; outline-color: rgba(231, 234, 243, 0.3); border-color: #141927; background-image: image(#151b2a); box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.incognito-mode headerbar button:disabled, .incognito-mode actionbar button:disabled { color: #868d9e; border-color: #141927; background-image: image(#252f49); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.incognito-mode headerbar button:backdrop, .incognito-mode actionbar button:backdrop { color: #868d9e; border-color: #171d2e; background-image: image(#252f49); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.incognito-mode headerbar button:backdrop:active, .incognito-mode actionbar button:backdrop:active { color: #868d9e; border-color: #171d2e; background-image: image(#1e263b); box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.incognito-mode headerbar button:backdrop:disabled, .incognito-mode actionbar button:backdrop:disabled { color: #3e507c; border-color: #171d2e; background-image: image(#252f49); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.incognito-mode headerbar button:checked, .incognito-mode actionbar button:checked { background: image(#151b2a); border-color: #0e131d; border-top-color: #06080c; }

.incognito-mode headerbar button:checked:backdrop, .incognito-mode actionbar button:checked:backdrop { color: #868d9e; border-color: #171d2e; background-image: image(#1e263b); box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.incognito-mode headerbar button.flat:backdrop, .incognito-mode headerbar button.titlebutton:backdrop, .incognito-mode actionbar button.flat:backdrop, .incognito-mode actionbar button.titlebutton:backdrop { border-color: transparent; background-color: transparent; background-image: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.incognito-mode headerbar entry, .incognito-mode actionbar entry { border-color: #141927; }

.incognito-mode headerbar entry:focus, .incognito-mode actionbar entry:focus { border-color: #030c17; box-shadow: inset 0 0 0 1px #15539e; }

.incognito-mode headerbar entry:backdrop, .incognito-mode actionbar entry:backdrop { box-shadow: none; border-color: #171d2e; }

.incognito-mode tabbar .box { background: #151b2a; border-bottom: 1px solid #06080c; }

.incognito-mode tabbar .box:backdrop { background-color: #151b2a; border-color: #171d2e; }

.incognito-mode tabbar scrolledwindow.pinned undershoot { border-color: #06080c; }

.incognito-mode tabbar scrolledwindow.pinned:backdrop undershoot { border-color: #171d2e; }

.incognito-mode tabbar scrolledwindow.pinned tabbox:dir(ltr) { box-shadow: inset -1px 0 #06080c; }

.incognito-mode tabbar scrolledwindow.pinned tabbox:dir(ltr):backdrop { box-shadow: inset -1px 0 #171d2e; }

.incognito-mode tabbar scrolledwindow.pinned tabbox:dir(rtl) { box-shadow: inset 1px 0 #06080c; }

.incognito-mode tabbar scrolledwindow.pinned tabbox:dir(rtl):backdrop { box-shadow: inset 1px 0 #171d2e; }

.incognito-mode tabbar tab { border-color: #06080c; background-color: #1a2235; }

.incognito-mode tabbar tab:checked { background-color: #252f49; }

.incognito-mode tabbar tab:checked:hover { background-color: #2a3553; }

.incognito-mode tabbar tab:hover { background-color: #1f283f; }

.incognito-mode tabbar tab:backdrop { border-color: #171d2e; background-color: #1a2235; }

.incognito-mode tabbar tab:backdrop:checked { background-color: #252f49; }

.incognito-mode tabbar tab .tab-close-button:hover, .incognito-mode tabbar tab .tab-indicator.clickable:hover { background: alpha(#e7eaf3,0.15); }

.incognito-mode tabbar tab .tab-close-button:active, .incognito-mode tabbar tab .tab-indicator.clickable:active { background: alpha(black,0.2); }

.incognito-mode tabbar .start-action, .incognito-mode tabbar .end-action { background: #1a2235; border-color: #06080c; }

.incognito-mode tabbar .start-action:backdrop, .incognito-mode tabbar .end-action:backdrop { border-color: #171d2e; background-color: #1a2235; }

.automation-mode { /*************************** Check and Radio buttons * */ }

.automation-mode headerbar { background: #e68700 linear-gradient(to top, #fa9300, #ff9805); box-shadow: inset 0 1px rgba(255, 240, 219, 0.07); border-color: #bd6f00; }

.automation-mode headerbar > clamp { color: #fff0db; }

.automation-mode headerbar:backdrop { background-image: image(#ffa11a); border-color: #eb8c05; box-shadow: inset 0 1px rgba(255, 240, 219, 0.07); }

.automation-mode headerbar:backdrop > clamp { color: #ffc87a; }

.automation-mode actionbar { background: #ffa11a; }

.automation-mode actionbar > revealer > box { border-color: #e68700; }

.automation-mode actionbar:backdrop { background: #ffa11a; }

.automation-mode actionbar:backdrop > revealer > box { border-color: #eb8c05; }

.automation-mode headerbar > *, .automation-mode actionbar > * { color: #fff0db; }

.automation-mode headerbar > *:backdrop, .automation-mode actionbar > *:backdrop { color: #ffc87a; }

.automation-mode headerbar button, .automation-mode actionbar button { color: #fff0db; outline-color: rgba(255, 240, 219, 0.3); border-color: #e68700; border-bottom-color: #bd6f00; background-image: linear-gradient(to top, #ff9e14 2px, #ffa11a); text-shadow: 0 -1px rgba(0, 0, 0, 0.56); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.56); box-shadow: inset 0 1px rgba(255, 255, 255, 0.2), 0 1px 2px rgba(0, 0, 0, 0.07); }

.automation-mode headerbar button.flat, .automation-mode headerbar button.titlebutton, .automation-mode actionbar button.flat, .automation-mode actionbar button.titlebutton { border-color: transparent; background-color: transparent; background-image: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.automation-mode headerbar button.titlebutton, .automation-mode actionbar button.titlebutton { text-shadow: 0 -1px rgba(0, 0, 0, 0.56); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.56); }

.automation-mode headerbar button:hover, .automation-mode actionbar button:hover { color: #fff0db; outline-color: rgba(255, 240, 219, 0.3); border-color: #e68700; border-bottom-color: #bd6f00; background-image: linear-gradient(to top, #ff9e14, #ffa31f 1px); text-shadow: 0 -1px rgba(0, 0, 0, 0.512); -gtk-icon-shadow: 0 -1px rgba(0, 0, 0, 0.512); box-shadow: inset 0 1px rgba(255, 255, 255, 0.2), 0 1px 2px rgba(0, 0, 0, 0.07); }

.automation-mode headerbar button:active, .automation-mode headerbar button:checked, .automation-mode actionbar button:active, .automation-mode actionbar button:checked { color: #fff0db; outline-color: rgba(255, 240, 219, 0.3); border-color: #e68700; background-image: image(#eb8a00); box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.automation-mode headerbar button:disabled, .automation-mode actionbar button:disabled { color: #ffc87a; border-color: #e68700; background-image: image(#ffa11a); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.automation-mode headerbar button:backdrop, .automation-mode actionbar button:backdrop { color: #ffc87a; border-color: #eb8c05; background-image: image(#ffa11a); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.automation-mode headerbar button:backdrop:active, .automation-mode actionbar button:backdrop:active { color: #ffc87a; border-color: #eb8c05; background-image: image(#ff9906); box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.automation-mode headerbar button:backdrop:disabled, .automation-mode actionbar button:backdrop:disabled { color: #ffc066; border-color: #eb8c05; background-image: image(#ffa11a); text-shadow: none; -gtk-icon-shadow: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.automation-mode headerbar button:checked, .automation-mode actionbar button:checked { background: image(#eb8a00); border-color: #d67e00; border-top-color: #bd6f00; }

.automation-mode headerbar button:checked:backdrop, .automation-mode actionbar button:checked:backdrop { color: #ffc87a; border-color: #eb8c05; background-image: image(#ff9906); box-shadow: inset 0 1px rgba(255, 255, 255, 0); }

.automation-mode headerbar button.flat:backdrop, .automation-mode headerbar button.titlebutton:backdrop, .automation-mode actionbar button.flat:backdrop, .automation-mode actionbar button.titlebutton:backdrop { border-color: transparent; background-color: transparent; background-image: none; box-shadow: inset 0 1px rgba(255, 255, 255, 0); text-shadow: none; -gtk-icon-shadow: none; }

.automation-mode headerbar entry, .automation-mode actionbar entry { border-color: #e68700; }

.automation-mode headerbar entry:focus, .automation-mode actionbar entry:focus { border-color: #030c17; box-shadow: inset 0 0 0 1px #15539e; }

.automation-mode headerbar entry:backdrop, .automation-mode actionbar entry:backdrop { box-shadow: none; border-color: #eb8c05; }

.automation-mode tabbar .box { background: #eb8a00; border-bottom: 1px solid #bd6f00; }

.automation-mode tabbar .box:backdrop { background-color: #eb8a00; border-color: #eb8c05; }

.automation-mode tabbar scrolledwindow.pinned undershoot { border-color: #bd6f00; }

.automation-mode tabbar scrolledwindow.pinned:backdrop undershoot { border-color: #eb8c05; }

.automation-mode tabbar scrolledwindow.pinned tabbox:dir(ltr) { box-shadow: inset -1px 0 #bd6f00; }

.automation-mode tabbar scrolledwindow.pinned tabbox:dir(ltr):backdrop { box-shadow: inset -1px 0 #eb8c05; }

.automation-mode tabbar scrolledwindow.pinned tabbox:dir(rtl) { box-shadow: inset 1px 0 #bd6f00; }

.automation-mode tabbar scrolledwindow.pinned tabbox:dir(rtl):backdrop { box-shadow: inset 1px 0 #eb8c05; }

.automation-mode tabbar tab { border-color: #bd6f00; background-color: #fa9300; }

.automation-mode tabbar tab:checked { background-color: #ffa11a; }

.automation-mode tabbar tab:checked:hover { background-color: #ffa729; }

.automation-mode tabbar tab:hover { background-color: #ff9a0a; }

.automation-mode tabbar tab:backdrop { border-color: #eb8c05; background-color: #fa9300; }

.automation-mode tabbar tab:backdrop:checked { background-color: #ffa11a; }

.automation-mode tabbar tab .tab-close-button:hover, .automation-mode tabbar tab .tab-indicator.clickable:hover { background: alpha(#fff0db,0.15); }

.automation-mode tabbar tab .tab-close-button:active, .automation-mode tabbar tab .tab-indicator.clickable:active { background: alpha(black,0.2); }

.automation-mode tabbar .start-action, .automation-mode tabbar .end-action { background: #fa9300; border-color: #bd6f00; }

.automation-mode tabbar .start-action:backdrop, .automation-mode tabbar .end-action:backdrop { border-color: #eb8c05; background-color: #fa9300; }

/*************************** Check and Radio buttons * */
.url_entry:not(:hover):not(:focus) { background-color: rgba(45, 45, 45, 0.75); }

.bookmark-tag-widget { border-radius: 5px; }

.bookmark-tag-widget button { background: none; border: none; box-shadow: none; -gtk-icon-shadow: none; text-shadow: none; min-width: 24px; min-height: 24px; border-radius: 99px; color: inherit; padding: 0; margin: 2px; }

.bookmark-tag-widget button:hover { background: alpha(#eeeeec,0.15); }

.bookmark-tag-widget button:hover:active { background: alpha(black,0.2); }

.bookmark-tag-widget-selected button { background: none; border: none; box-shadow: none; -gtk-icon-shadow: none; text-shadow: none; min-width: 24px; min-height: 24px; border-radius: 99px; color: inherit; padding: 0; margin: 2px; }

.bookmark-tag-widget-selected button:hover { background: alpha(#ffffff,0.15); }

.bookmark-tag-widget-selected button:hover:active { background: alpha(black,0.2); }

@keyframes needs_attention_keyframes { 0% { background-image: linear-gradient(to bottom, #fafafa, #ededed 40%, #e0e0e0);
    border-color: @borders; }
  /* can't do animation-direction, so holding the color on two keyframes */
  30% { background-image: linear-gradient(to bottom, @theme_base_color, @theme_base_color, @theme_base_color);
    border-color: @theme_fg_color; }
  90% { background-image: linear-gradient(to bottom, @theme_base_color, @theme_base_color, @theme_base_color);
    border-color: @theme_fg_color; }
  100% { background-image: linear-gradient(to bottom, #fafafa, #ededed 40%, #e0e0e0);
    border-color: @borders; } }

.epiphany-downloads-button-needs-attention { animation: needs_attention_keyframes 2s ease-in-out; }

.epiphany-downloads-button-needs-attention-multiple { animation: needs_attention_keyframes 3s ease-in-out; animation-iteration-count: 3; }

#title-box-container { color: alpha(@theme_fg_color,0.2); }

#title-box-container > clamp { color: @theme_fg_color; }

#title-box-container > clamp:backdrop { color: @theme_unfocused_fg_color; }

.pages-list { margin: 5px; }

.pages-list row { min-height: 30px; padding: 3px; border-radius: 5px; -gtk-outline-radius: 5px; border: none; color: #eeeeec; transition: none; box-shadow: none; }

.pages-list row:not(:last-child) { margin-bottom: 2px; }

.pages-list row:selected { background: #202020; }

.pages-list row:hover { background: #141414; }

.pages-list row .close-button { background: none; border: none; box-shadow: none; -gtk-icon-shadow: none; text-shadow: none; min-width: 24px; min-height: 24px; border-radius: 99px; color: inherit; padding: 0; margin: 2px; }

.pages-list row .close-button:hover { background: alpha(#eeeeec,0.15); }

.pages-list row .close-button:hover:active { background: alpha(black,0.2); }

#pages-view .pages-list row .close-button { min-width: 36px; min-height: 36px; }
